#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <allegro.h>
#include "edop.h"
#include "ymw.h"
#include "mem.h"
#include "find.h"

int new_world( WORLD *world )
{
    world->num_zones = 0;
    if( (world->zone = malloc( 1 )) == NULL )
      return 0;
    world->last_zone_vnum = -1;

    return -1;
}

int destroy_world( WORLD *world )
{
    for( ; world->num_zones > 0; ){
       del_zone( world, 0 );
       }
    free( world->zone );

    world->num_zones = 0;
    world->last_zone_vnum = -1;

    return -1;
}

int add_zone( WORLD *world )
{
    int j;

    j = world->num_zones++;
    world->zone = realloc( world->zone, sizeof(ZONE)*world->num_zones );

    world->zone[j].name = malloc( NAMESIZE );
    world->zone[j].author = malloc( NAMESIZE );
    memset( world->zone[j].name, 0, NAMESIZE );
    memset( world->zone[j].author, 0, NAMESIZE );

    world->zone[j].tint = makecol( 200, 225, 200 );
    world->zone[j].vnum = (++world->last_zone_vnum);
    world->zone[j].life_span = 15;
    world->zone[j].reset_mode = 2;
    world->zone[j].rbv = 0;
    world->zone[j].stype = 0;

    world->zone[j].num_rooms = 0;
    world->zone[j].room = malloc( 1 );
    world->zone[j].last_room_vnum = -1;

    world->zone[j].num_objects = 0;
    world->zone[j].object = malloc( 1 );
    world->zone[j].last_object_vnum = -1;

    world->zone[j].num_mobiles = 0;
    world->zone[j].mobile = malloc( 1 );
    world->zone[j].last_mobile_vnum = -1;

    world->zone[j].num_shops = 0;
    world->zone[j].shop = malloc( 1 );
    world->zone[j].last_shop_vnum = -1;

    world->zone[j].num_tags = 0;
    world->zone[j].tag = malloc( 1 );

    world->zone[j].parent = world;

    rehash_world( world );

    return j;
}

int del_zone( WORLD *world, int index )
{
    int j;

    free( world->zone[index].name );
    free( world->zone[index].author );

    for( ; world->zone[index].num_rooms > 0; ){
       del_room( &world->zone[index], 0 );
       }
    free( world->zone[index].room );

    for( ; world->zone[index].num_objects > 0; ){
       del_object( &world->zone[index], 0 );
       }
    free( world->zone[index].object );

    for( ; world->zone[index].num_mobiles > 0; ){
       del_mobile( &world->zone[index], 0 );
       }
    free( world->zone[index].mobile );

    for( ; world->zone[index].num_shops > 0; ){
       del_shop( &world->zone[index], 0 );
       }
    free( world->zone[index].shop );

    for( ; world->zone[index].num_tags > 0; ){
       del_tag( &world->zone[index], 0 );
       }
    free( world->zone[index].tag );

    for( j = index; j < world->num_zones-1; j++ ){
       world->zone[j].name             = world->zone[j+1].name;
       world->zone[j].author           = world->zone[j+1].author;
       world->zone[j].tint             = world->zone[j+1].tint;
       world->zone[j].vnum             = world->zone[j+1].vnum;
       world->zone[j].life_span        = world->zone[j+1].life_span;
       world->zone[j].reset_mode       = world->zone[j+1].reset_mode;
       world->zone[j].rbv              = world->zone[j+1].rbv;
       world->zone[j].stype            = world->zone[j+1].stype;

       world->zone[j].num_rooms        = world->zone[j+1].num_rooms;
       world->zone[j].room             = world->zone[j+1].room;
       world->zone[j].last_room_vnum   = world->zone[j+1].last_room_vnum;

       world->zone[j].num_objects      = world->zone[j+1].num_objects;
       world->zone[j].object           = world->zone[j+1].object;
       world->zone[j].last_object_vnum = world->zone[j+1].last_object_vnum;

       world->zone[j].num_mobiles      = world->zone[j+1].num_mobiles;
       world->zone[j].mobile           = world->zone[j+1].mobile;
       world->zone[j].last_mobile_vnum = world->zone[j+1].last_mobile_vnum;

       world->zone[j].num_shops        = world->zone[j+1].num_shops;
       world->zone[j].shop             = world->zone[j+1].shop;
       world->zone[j].last_shop_vnum   = world->zone[j+1].last_shop_vnum;

       world->zone[j].num_tags         = world->zone[j+1].num_tags;
       world->zone[j].tag              = world->zone[j+1].tag;
       }

    world->num_zones--;
    world->zone = realloc( world->zone, MAX(1, sizeof(ZONE)*world->num_zones) );

    rehash_world( world );

    return 0;
}

int add_room( ZONE *zone, int x, int y, int z )
{
    int j;
 
    j = zone->num_rooms;
    zone->num_rooms++;
 
    zone->room = realloc( zone->room, sizeof(ROOM)*zone->num_rooms );
    zone->room[j].x = x;
    zone->room[j].y = y;
    zone->room[j].z = z;
    zone->room[j].title = malloc( NAMESIZE );
    memset( zone->room[j].title, 0, NAMESIZE );
 
    zone->room[j].desc = malloc( RDSIZE );
    memset( zone->room[j].desc, 0, RDSIZE );

    zone->room[j].vnum = (++zone->last_room_vnum);
    zone->room[j].rbv = zone->rbv;
    zone->room[j].stype = zone->stype;
 
    zone->room[j].num_doors = 0;
    zone->room[j].door = malloc( 1 );
 
    zone->room[j].num_exd = 0;
    zone->room[j].exd = malloc( 1 );
 
    zone->room[j].num_cont = 0;
    zone->room[j].content = malloc( 1 );
 
    zone->room[j].parent = zone;

    rehash_world( zone->parent );
 
    return j;
}

int del_room( ZONE *zone, int index )
{
    int j;
 
    for( ; zone->room[index].num_doors > 0; ){
       del_door( &zone->room[index], 0 );
       }
    for( j = 0; j < zone->room[index].num_exd; j++ ){
       free( zone->room[index].exd[j].keywords );
       free( zone->room[index].exd[j].desc );
       }
 
    free( zone->room[index].door );
    free( zone->room[index].title );
    free( zone->room[index].desc );
    free( zone->room[index].exd );
    free( zone->room[index].content );
    for( j = index; j < zone->num_rooms-1; j++ ){
       zone->room[j].x         = zone->room[j+1].x;
       zone->room[j].y         = zone->room[j+1].y;
       zone->room[j].z         = zone->room[j+1].z;
       zone->room[j].title     = zone->room[j+1].title;
       zone->room[j].desc      = zone->room[j+1].desc;
       zone->room[j].vnum      = zone->room[j+1].vnum;
       zone->room[j].zone      = zone->room[j+1].zone;
       zone->room[j].rbv       = zone->room[j+1].rbv;
       zone->room[j].stype     = zone->room[j+1].stype;

       zone->room[j].num_doors = zone->room[j+1].num_doors;
       zone->room[j].door      = zone->room[j+1].door;

       zone->room[j].num_exd   = zone->room[j+1].num_exd;
       zone->room[j].exd       = zone->room[j+1].exd;

       zone->room[j].num_cont  = zone->room[j+1].num_cont;
       zone->room[j].content   = zone->room[j+1].content;

       zone->room[j].parent    = zone->room[j+1].parent;
       }
 
    zone->num_rooms--;
    zone->room = realloc( zone->room, MAX(sizeof(ROOM)*zone->num_rooms,1) );

    rehash_world( zone->parent );

    return 0;
}

int move_room_to_zone( ZONE *from, int index, ZONE *to )
{
    int id2, j;

    id2 = add_room( to, from->room[index].x, from->room[index].y, from->room[index].z );

    free( to->room[id2].title );
    free( to->room[id2].desc );
    free( to->room[id2].door );
    free( to->room[id2].exd );
    free( to->room[id2].content );

    to->room[id2].title     = from->room[index].title;
    to->room[id2].desc      = from->room[index].desc;
    to->room[id2].zone      = from->room[index].zone;
    to->room[id2].rbv       = from->room[index].rbv;
    to->room[id2].stype     = from->room[index].stype;
    to->room[id2].num_doors = from->room[index].num_doors;
    to->room[id2].door      = from->room[index].door;
    to->room[id2].num_exd   = from->room[index].num_exd;
    to->room[id2].exd       = from->room[index].exd;
    to->room[id2].num_cont  = from->room[index].num_cont;
    to->room[id2].content   = from->room[index].content;

    from->room[index].title = NULL;
    from->room[index].desc = NULL;
    from->room[index].num_doors = 0;
    from->room[index].door = NULL;
    from->room[index].num_exd = 0;
    from->room[index].exd = NULL;
    from->room[index].num_cont = 0;
    from->room[index].content = NULL;

    for( j = 0; j < to->room[id2].num_doors; j++ ){
       to->room[id2].door[j].other->other = &to->room[id2].door[j];
       to->room[id2].door[j].other->croom = to->room[id2].vnum;
       to->room[id2].door[j].other->czone = to->vnum;
       }

    del_room( from, index );

    return id2;
}

int add_room_exd( ROOM *room )
{
    int j;

    j = room->num_exd++;
    room->exd = realloc( room->exd, sizeof(EXD)*room->num_exd );

    room->exd[j].keywords = malloc( NAMESIZE );
    memset( room->exd[j].keywords, 0, NAMESIZE );

    room->exd[j].desc = malloc( RDSIZE );
    memset( room->exd[j].desc, 0, RDSIZE );

    return j;
}

int del_room_exd( ROOM *room, int index )
{
    int j;

    free( room->exd[index].keywords );
    free( room->exd[index].desc );

    for( j = index; j < room->num_exd-1; j++ ){
       room->exd[j].keywords = room->exd[j+1].keywords;
       room->exd[j].desc     = room->exd[j+1].desc;
       }

    room->num_exd--;
    room->exd = realloc( room->exd, MAX( 1, sizeof(EXD)*room->num_exd) );

    return 0;
}

int add_room_content( ROOM *room )
{
    int j;

    j = room->num_cont++;
    room->content = realloc( room->content, sizeof(RCONTENT)*room->num_cont );

    room->content[j].type = 0;
    room->content[j].vnum = 0;
    room->content[j].ifflag = 0;
    room->content[j].max = 0;
    room->content[j].extra = 0;

    return j;
}

int del_room_content( ROOM *room, int index )
{
    int j;

    for( j = index; j < room->num_cont-1; j++ ){
       room->content[j].type   = room->content[j+1].type;
       room->content[j].vnum   = room->content[j+1].vnum;
       room->content[j].ifflag = room->content[j+1].ifflag;
       room->content[j].max    = room->content[j+1].max;
       room->content[j].extra  = room->content[j+1].extra;
       }

    room->num_cont--;
    room->content = realloc( room->content, sizeof(RCONTENT)*room->num_cont );

    return 0;
}

int add_object_exd( OBJECT *object )
{
    int j;

    j = object->num_exd++;
    object->exd = realloc( object->exd, sizeof(EXD)*object->num_exd );

    object->exd[j].keywords = malloc( NAMESIZE );
    memset( object->exd[j].keywords, 0, NAMESIZE );

    object->exd[j].desc = malloc( RDSIZE );
    memset( object->exd[j].desc, 0, RDSIZE );

    return j;
}

int del_object_exd( OBJECT *object, int index )
{
    int j;

    free( object->exd[index].keywords );
    free( object->exd[index].desc );

    for( j = index; j < object->num_exd-1; j++ ){
       object->exd[j].keywords = object->exd[j+1].keywords;
       object->exd[j].desc     = object->exd[j+1].desc;
       }

    object->num_exd--;
    object->exd = realloc( object->exd, MAX( 1, sizeof(EXD)*object->num_exd) );

    return 0;
}

int add_object_affect( OBJECT *object )
{
    int j/*, k*/;

    j = object->num_affect++;
    object->affect = realloc( object->affect, sizeof(EXD)*object->num_affect );

    object->affect[j].location = 0;
    object->affect[j].value = 0;

    return j;
}

int del_object_affect( OBJECT *object, int index )
{
    int j;

    for( j = index; j < object->num_affect-1; j++ ){
       object->affect[j].location = object->affect[j+1].location;
       object->affect[j].value    = object->affect[j+1].value;
       }

    object->num_affect--;
    object->affect = realloc( object->affect, MAX( 1, sizeof(EXD)*object->num_affect) );

    return 0;
}

int add_door( ROOM *room1, int dir1, ROOM *room2, int dir2 )
{
    int j1, j2;

    if( room1->num_doors > ND ||
        room2->num_doors > ND ){
      return -1;
      }

    j1 = room1->num_doors;
    room1->num_doors++;
    room1->door = realloc( room1->door, sizeof(DOOR)*room1->num_doors );

    room1->door[j1].dir   = dir1;
    room1->door[j1].cdir  = dir2;
    room1->door[j1].croom = room2->vnum;
    room1->door[j1].czone = room2->parent->vnum;

    room1->door[j1].desc = malloc( RDSIZE );
    memset( room1->door[j1].desc, 0, RDSIZE );

    room1->door[j1].keywords = malloc( NAMESIZE );
    memset( room1->door[j1].keywords, 0, NAMESIZE );

    room1->door[j1].typea = 0;
    room1->door[j1].typeb = 0;
    room1->door[j1].state = 0;
    room1->door[j1].key = -1;

    room1->door[j1].parent = room1;

    j2 = room2->num_doors;
    room2->num_doors++;
    room2->door = realloc( room2->door, sizeof(DOOR)*room2->num_doors );

    room2->door[j2].dir   = dir2;
    room2->door[j2].cdir  = dir1;
    room2->door[j2].croom = room1->vnum;
    room2->door[j2].czone = room1->parent->vnum;

    room2->door[j2].desc = malloc( RDSIZE );
    memset( room2->door[j2].desc, 0, RDSIZE );

    room2->door[j2].keywords = malloc( NAMESIZE );
    memset( room2->door[j2].keywords, 0, NAMESIZE );

    room2->door[j2].typea = 0;
    room2->door[j2].typeb = 0;
    room2->door[j2].state = 0;
    room2->door[j2].key = -1;

    room2->door[j2].parent = room2;

    room1->door[j1].other = &room2->door[j2];
    room2->door[j2].other = &room1->door[j1];

    rehash_world( room1->parent->parent );

    return 0;
}

int add_hanging_door( ROOM *room, int dir )
{
    int j;

    if( room->num_doors > ND ){
      return -1;
      }

    j = room->num_doors;
    room->num_doors++;
    room->door = realloc( room->door, sizeof(DOOR)*room->num_doors );

    room->door[j].dir   = dir;
    room->door[j].cdir  = -1;
    room->door[j].croom = -1;
    room->door[j].czone = -1;

    room->door[j].desc = malloc( RDSIZE );
    memset( room->door[j].desc, 0, RDSIZE );

    room->door[j].keywords = malloc( NAMESIZE );
    memset( room->door[j].keywords, 0, NAMESIZE );

    room->door[j].typea = 0;
    room->door[j].typeb = 0;
    room->door[j].state = 0;
    room->door[j].key = -1;

    room->door[j].parent = room;

    room->door[j].other = NULL;

    return j;
}

int del_door( ROOM *room, int index )
{
    int j, index2;
    ROOM *room2;

    if( !room->door[index].other ){
      room2 = NULL;
      index2 = -1;
      }
    else{
      room2 = room->door[index].other->parent;
      index2 = ifind_door_dir_room( room2, room->door[index].cdir );
      }

    free( room->door[index].desc );
    free( room->door[index].keywords );

    for( j = index; j < room->num_doors-1; j++ ){
       room->door[j].croom      = room->door[j+1].croom;
       room->door[j].czone      = room->door[j+1].czone;
       room->door[j].cdir       = room->door[j+1].cdir;
       room->door[j].typea      = room->door[j+1].typea;
       room->door[j].typeb      = room->door[j+1].typeb;
       room->door[j].state      = room->door[j+1].state;
       room->door[j].key        = room->door[j+1].key;
       room->door[j].dir        = room->door[j+1].dir;
       room->door[j].desc       = room->door[j+1].desc;
       room->door[j].keywords   = room->door[j+1].keywords;
       room->door[j].other      = room->door[j+1].other;
       room->door[j].parent     = room->door[j+1].parent;
       }

    if( room2 ){
      free( room2->door[index2].desc );
      free( room2->door[index2].keywords );
  
      for( j = index2; j < room2->num_doors-1; j++ ){
         room2->door[j].croom      = room2->door[j+1].croom;
         room2->door[j].czone      = room2->door[j+1].czone;
         room2->door[j].cdir       = room2->door[j+1].cdir;
         room2->door[j].typea      = room2->door[j+1].typea;
         room2->door[j].typeb      = room2->door[j+1].typeb;
         room2->door[j].state      = room2->door[j+1].state;
         room2->door[j].key        = room2->door[j+1].key;
         room2->door[j].dir        = room2->door[j+1].dir;
         room2->door[j].desc       = room2->door[j+1].desc;
         room2->door[j].keywords   = room2->door[j+1].keywords;
         room2->door[j].other      = room2->door[j+1].other;
         room2->door[j].parent     = room2->door[j+1].parent;
         }

      room2->num_doors--;
      room2->door = realloc( room2->door, sizeof(ROOM)*room2->num_doors );
      }
    else{
      room2 = room;
      room2 = NULL;
      }

    room->num_doors--;
    room->door = realloc( room->door, sizeof(ROOM)*room->num_doors );

    rehash_world( room->parent->parent );

    return 0;
}

int add_tag( ZONE *zone, int x, int y, int z )
{
    int j, k;

    j = zone->num_tags++;
    zone->tag = realloc( zone->tag, sizeof(TAG)*zone->num_tags );

    zone->tag[j].x = x;
    zone->tag[j].y = y;
    zone->tag[j].z = z;
    zone->tag[j].str = malloc( NAMESIZE );
    for( k = 0; k < NAMESIZE/8; k++ ){
       ((long long *)zone->tag[j].str)[k] = 0;
       }

    zone->tag[j].parent = zone;

    rehash_world( zone->parent );

    return j;
}

int del_tag( ZONE *zone, int index )
{
    int j;

    free( zone->tag[index].str );

    for( j = index; j < zone->num_tags-1; j++ ){
       zone->tag[j].x      = zone->tag[j+1].x;
       zone->tag[j].y      = zone->tag[j+1].y;
       zone->tag[j].z      = zone->tag[j+1].z;
       zone->tag[j].str    = zone->tag[j+1].str;
       zone->tag[j].parent = zone->tag[j+1].parent;
       }

    zone->num_tags--;
    zone->tag = realloc( zone->tag, MAX(1,sizeof(TAG)*zone->num_tags) );

    rehash_world( zone->parent );

    return 0;
}

int add_object( ZONE *zone )
{
    int j, k;

    j = zone->num_objects++;
    zone->object = realloc( zone->object, sizeof(OBJECT)*zone->num_objects );
    zone->object[j].vnum = (++zone->last_object_vnum);
    zone->object[j].keywords = malloc( NAMESIZE );
    for( k = 0; k < NAMESIZE/8; k++ ){
       ((long long *)zone->object[j].keywords)[k] = 0;
       }

    zone->object[j].short_desc = malloc( NAMESIZE );
    for( k = 0; k < NAMESIZE/8; k++ ){
       ((long long *)zone->object[j].short_desc)[k] = 0;
       }

    zone->object[j].long_desc = malloc( RDSIZE );
    for( k = 0; k < RDSIZE/8; k++ ){
       ((long long *)zone->object[j].long_desc)[k] = 0;
       }

    zone->object[j].action_desc = malloc( RDSIZE );
    for( k = 0; k < RDSIZE/8; k++ ){
       ((long long *)zone->object[j].action_desc)[k] = 0;
       }

    zone->object[j].type = 0;
    zone->object[j].ebv = 0;
    zone->object[j].wbv = 0;
    for( k = 0; k < 4; k++ )
       zone->object[j].value[k] = 0;
    zone->object[j].weight = 0;
    zone->object[j].cost = 0;
    zone->object[j].rpd = 0;
    zone->object[j].num_exd = 0;
    zone->object[j].exd = malloc( 1 );
    zone->object[j].num_affect = 0;
    zone->object[j].affect = malloc( 1 );
    zone->object[j].parent = zone;

    rehash_world( zone->parent );

    return j;
}

int del_object( ZONE *zone, int index )
{
    int j, k;

    free( zone->object[index].keywords );
    free( zone->object[index].short_desc );
    free( zone->object[index].long_desc );
    free( zone->object[index].action_desc );
    free( zone->object[index].affect );
    for( ; zone->object[index].num_exd > 0; ){
       del_object_exd( &zone->object[index], 0 );
       }

    for( j = index; j < zone->num_objects-1; j++ ){
       zone->object[j].vnum        = zone->object[j+1].vnum;
       zone->object[j].keywords    = zone->object[j+1].keywords;
       zone->object[j].short_desc  = zone->object[j+1].short_desc;
       zone->object[j].long_desc   = zone->object[j+1].long_desc;
       zone->object[j].action_desc = zone->object[j+1].action_desc;
       zone->object[j].type        = zone->object[j+1].type;
       zone->object[j].ebv         = zone->object[j+1].ebv;
       zone->object[j].wbv         = zone->object[j+1].wbv;
       for( k = 0; k < 4; k++ ){
          zone->object[j].value[k] = zone->object[j+1].value[k];
          }
       zone->object[j].weight      = zone->object[j+1].weight;
       zone->object[j].cost        = zone->object[j+1].cost;
       zone->object[j].rpd         = zone->object[j+1].rpd;
       zone->object[j].num_exd     = zone->object[j+1].num_exd;
       zone->object[j].exd         = zone->object[j+1].exd;
       zone->object[j].num_affect  = zone->object[j+1].num_affect;
       zone->object[j].affect      = zone->object[j+1].affect;
       zone->object[j].parent      = zone->object[j+1].parent;
       }

    zone->num_objects--;
    zone->object = realloc( zone->object, MAX(1,sizeof(OBJECT)*zone->num_objects) );

    rehash_world( zone->parent );

    return 0;
}

int add_mobile( ZONE *zone )
{
    int j, k;

    j = zone->num_mobiles++;
    zone->mobile = realloc( zone->mobile, sizeof(MOBILE)*zone->num_mobiles );
    zone->mobile[j].vnum = (++zone->last_mobile_vnum);
    zone->mobile[j].keywords = malloc( NAMESIZE );
    for( k = 0; k < NAMESIZE/8; k++ ){
       ((long long *)zone->mobile[j].keywords)[k] = 0;
       }

    zone->mobile[j].short_desc = malloc( NAMESIZE );
    for( k = 0; k < NAMESIZE/8; k++ ){
       ((long long *)zone->mobile[j].short_desc)[k] = 0;
       }

    zone->mobile[j].long_desc = malloc( NAMESIZE );
    for( k = 0; k < NAMESIZE/8; k++ ){
       ((long long *)zone->mobile[j].long_desc)[k] = 0;
       }

    zone->mobile[j].detail_desc = malloc( RDSIZE );
    for( k = 0; k < RDSIZE/8; k++ ){
       ((long long *)zone->mobile[j].detail_desc)[k] = 0;
       }

    zone->mobile[j].actbv = 0;
    zone->mobile[j].affbv = 0;
    zone->mobile[j].alignment = 0;
    zone->mobile[j].typeflag = 0;
    zone->mobile[j].level = 0;
    zone->mobile[j].thac0 = 0;
    zone->mobile[j].ac = 0;
    zone->mobile[j].mhpx = 0;
    zone->mobile[j].mhpy = 0;
    zone->mobile[j].mhpz = 0;
    zone->mobile[j].bhdx = 0;
    zone->mobile[j].bhdy = 0;
    zone->mobile[j].bhdz = 0;
    zone->mobile[j].gold = 0;
    zone->mobile[j].exp = 0;
    zone->mobile[j].loadpos = 0;
    zone->mobile[j].defpos = 0;
    zone->mobile[j].sex = 0;
    zone->mobile[j].num_especs = 0;
    zone->mobile[j].espec = malloc( 1 );
    zone->mobile[j].parent = zone;

    rehash_world( zone->parent );

    return j;
}

int del_mobile( ZONE *zone, int index )
{
    int j;

    free( zone->mobile[index].short_desc );
    free( zone->mobile[index].long_desc );
    free( zone->mobile[index].detail_desc );
    free( zone->mobile[index].keywords );
    free( zone->mobile[index].espec );

    for( j = index; j < zone->num_mobiles-1; j++ ){
       zone->mobile[j].vnum        = zone->mobile[j+1].vnum;
       zone->mobile[j].keywords    = zone->mobile[j+1].keywords;
       zone->mobile[j].short_desc  = zone->mobile[j+1].short_desc;
       zone->mobile[j].long_desc   = zone->mobile[j+1].long_desc;
       zone->mobile[j].detail_desc = zone->mobile[j+1].detail_desc;
       zone->mobile[j].actbv       = zone->mobile[j+1].actbv;
       zone->mobile[j].affbv       = zone->mobile[j+1].affbv;
       zone->mobile[j].alignment   = zone->mobile[j+1].alignment;
       zone->mobile[j].typeflag    = zone->mobile[j+1].typeflag;
       zone->mobile[j].level       = zone->mobile[j+1].level;
       zone->mobile[j].thac0       = zone->mobile[j+1].thac0;
       zone->mobile[j].ac          = zone->mobile[j+1].ac;
       zone->mobile[j].mhpx        = zone->mobile[j+1].mhpx;
       zone->mobile[j].mhpy        = zone->mobile[j+1].mhpy;
       zone->mobile[j].mhpz        = zone->mobile[j+1].mhpz;
       zone->mobile[j].bhdx        = zone->mobile[j+1].bhdx;
       zone->mobile[j].bhdy        = zone->mobile[j+1].bhdy;
       zone->mobile[j].bhdz        = zone->mobile[j+1].bhdz;
       zone->mobile[j].gold        = zone->mobile[j+1].gold;
       zone->mobile[j].exp         = zone->mobile[j+1].exp;
       zone->mobile[j].loadpos     = zone->mobile[j+1].loadpos;
       zone->mobile[j].defpos      = zone->mobile[j+1].defpos;
       zone->mobile[j].sex         = zone->mobile[j+1].sex;
       zone->mobile[j].num_especs  = zone->mobile[j+1].num_especs;
       zone->mobile[j].espec       = zone->mobile[j+1].espec;
       zone->mobile[j].parent      = zone->mobile[j+1].parent;
       }

    zone->num_mobiles--;
    zone->mobile = realloc( zone->mobile, MAX(1,sizeof(MOBILE)*zone->num_mobiles) );

    rehash_world( zone->parent );

    return 0;
}

int add_shop( ZONE *zone )
{
    int j;

    j = zone->num_shops++;
    zone->shop = realloc( zone->shop, sizeof(SHOP)*zone->num_shops );

    zone->shop[j].vnum = (++zone->last_shop_vnum);

    zone->shop[j].num_items = 0;
    zone->shop[j].item = malloc( 1 );

    zone->shop[j].prof_sell = 0.0;
    zone->shop[j].prof_buy = 0.0;

    zone->shop[j].num_bspecs = 0;
    zone->shop[j].bspec = malloc( 1 );

    zone->shop[j].buy_item_ne = malloc( NAMESIZE );
    memset( zone->shop[j].buy_item_ne, 0, NAMESIZE );

    zone->shop[j].sell_item_ne = malloc( NAMESIZE );
    memset( zone->shop[j].sell_item_ne, 0, NAMESIZE );

    zone->shop[j].refuse_offer = malloc( NAMESIZE );
    memset( zone->shop[j].refuse_offer, 0, NAMESIZE );

    zone->shop[j].shop_afford = malloc( NAMESIZE );
    memset( zone->shop[j].shop_afford, 0, NAMESIZE );

    zone->shop[j].play_afford = malloc( NAMESIZE );
    memset( zone->shop[j].play_afford, 0, NAMESIZE );

    zone->shop[j].buy_success = malloc( NAMESIZE );
    memset( zone->shop[j].buy_success, 0, NAMESIZE );

    zone->shop[j].sell_success = malloc( NAMESIZE );
    memset( zone->shop[j].sell_success, 0, NAMESIZE );

    zone->shop[j].temper = 0;
    zone->shop[j].shop_bv = 0;
    zone->shop[j].mobile = -1;
    zone->shop[j].withwho_bv = 0;

    zone->shop[j].num_rooms = 0;
    zone->shop[j].room = malloc( 1 );

    zone->shop[j].topen[0] = 0;    /* Setting these values like this means open all the time */
    zone->shop[j].tclose[0] = 28;
    zone->shop[j].topen[1] = 0;
    zone->shop[j].tclose[1] = 0;

    zone->shop[j].parent = zone;

    rehash_world( zone->parent );

    return j;
}

int del_shop( ZONE *zone, int index )
{
    int j;

    free( zone->shop[index].item );
    for( j = 0; j < zone->shop[index].num_bspecs; j++ ){
       free( zone->shop[index].bspec[j].namelist );
       }
    free( zone->shop[index].bspec );

    free( zone->shop[index].buy_item_ne );
    free( zone->shop[index].sell_item_ne );
    free( zone->shop[index].refuse_offer );
    free( zone->shop[index].shop_afford );
    free( zone->shop[index].play_afford );
    free( zone->shop[index].buy_success );
    free( zone->shop[index].sell_success );

    free( zone->shop[index].room );

    for( j = index; j < zone->num_shops-1; j++ ){
       zone->shop[j].vnum         = zone->shop[j+1].vnum;
       zone->shop[j].num_items    = zone->shop[j+1].num_items;
       zone->shop[j].item         = zone->shop[j+1].item;
       zone->shop[j].prof_sell    = zone->shop[j+1].prof_sell;
       zone->shop[j].prof_buy     = zone->shop[j+1].prof_buy;
       zone->shop[j].num_bspecs   = zone->shop[j+1].num_bspecs;
       zone->shop[j].bspec        = zone->shop[j+1].bspec;
       zone->shop[j].buy_item_ne  = zone->shop[j+1].buy_item_ne;
       zone->shop[j].sell_item_ne = zone->shop[j+1].sell_item_ne;
       zone->shop[j].refuse_offer = zone->shop[j+1].refuse_offer;
       zone->shop[j].shop_afford  = zone->shop[j+1].shop_afford;
       zone->shop[j].play_afford  = zone->shop[j+1].play_afford;
       zone->shop[j].buy_success  = zone->shop[j+1].buy_success;
       zone->shop[j].sell_success = zone->shop[j+1].sell_success;
       zone->shop[j].temper       = zone->shop[j+1].temper;
       zone->shop[j].shop_bv      = zone->shop[j+1].shop_bv;
       zone->shop[j].mobile       = zone->shop[j+1].mobile;
       zone->shop[j].withwho_bv   = zone->shop[j+1].withwho_bv;
       zone->shop[j].num_rooms    = zone->shop[j+1].num_rooms;
       zone->shop[j].room         = zone->shop[j+1].room;
       zone->shop[j].topen[0]     = zone->shop[j+1].topen[0];
       zone->shop[j].tclose[0]    = zone->shop[j+1].tclose[0];
       zone->shop[j].topen[1]     = zone->shop[j+1].topen[1];
       zone->shop[j].tclose[1]    = zone->shop[j+1].tclose[1];
       zone->shop[j].parent       = zone->shop[j+1].parent;
       }

    zone->num_shops--;
    zone->shop = realloc( zone->shop, sizeof(SHOP) * zone->num_shops );

    rehash_world( zone->parent );

    return zone->num_shops;
}

int add_shop_item( SHOP *shop )
{
    int j;

    j = shop->num_items++;
    shop->item = realloc( shop->item, sizeof(int)*shop->num_items );

    shop->item[j] = -1;

    return j;
}

int del_shop_item( SHOP *shop, int index )
{
    int j;

    for( j = index; j < shop->num_items-1; j++ ){
       shop->item[j] = shop->item[j+1];
       }

    shop->num_items--;
    shop->item = realloc( shop->item, sizeof(int)*shop->num_items );

    return shop->num_items;
}

int add_shop_room( SHOP *shop )
{
    int j;

    j = shop->num_rooms++;
    shop->room = realloc( shop->room, sizeof(int)*shop->num_rooms );

    shop->room[j] = -1;

    return j;
}

int del_shop_room( SHOP *shop, int index )
{
    int j;

    for( j = index; j < shop->num_rooms-1; j++ ){
       shop->room[j] = shop->room[j+1];
       }

    shop->num_rooms--;
    shop->room = realloc( shop->room, sizeof(int)*shop->num_rooms );

    return shop->num_rooms;
}

int add_shop_bspec( SHOP *shop )
{
    int j;

    j = shop->num_bspecs++;
    shop->bspec = realloc( shop->bspec, sizeof(BSPEC)*shop->num_bspecs );

    shop->bspec[j].type = 0;
    shop->bspec[j].namelist = malloc( NAMESIZE );
    memset( shop->bspec[j].namelist, 0, NAMESIZE );

    return j;
}

int del_shop_bspec( SHOP *shop, int index )
{
    int j;

    free( shop->bspec[index].namelist );

    for( j = index; j < shop->num_bspecs-1; j++ ){
       shop->bspec[j].type     = shop->bspec[j+1].type;
       shop->bspec[j].namelist = shop->bspec[j+1].namelist;
       }

    shop->num_bspecs--;
    shop->bspec = realloc( shop->bspec, sizeof(int)*shop->num_bspecs );

    return shop->num_bspecs;
}

void rehash_world( WORLD *world )
{
   int j, k;

   for( j = 0; j < world->num_zones; j++ ){
      world->zone[j].parent = world;
      rehash_zone( &world->zone[j] );
      }

   for( j = 0; j < world->num_zones; j++ ){
      for( k = 0; k < world->zone[j].num_rooms; k++ ){
         rehash_room( &world->zone[j].room[k] );
         }
      }
}

void rehash_zone( ZONE *zone )
{
    int j;

    for( j = 0; j < zone->num_rooms; j++ ){
       zone->room[j].parent = zone;
       }
    for( j = 0; j < zone->num_objects; j++ ){
       zone->object[j].parent = zone;
       }
    for( j = 0; j < zone->num_mobiles; j++ ){
       zone->mobile[j].parent = zone;
       }
    for( j = 0; j < zone->num_shops; j++ ){
       zone->shop[j].parent = zone;
       }
    for( j = 0; j < zone->num_tags; j++ ){
       zone->tag[j].parent = zone;
       }
}

void rehash_room( ROOM *room )
{
    int j;

    for( j = 0; j < room->num_doors; j++ ){
       room->door[j].parent = room;
       if( room->door[j].czone == -1 ||
           room->door[j].croom == -1 ||
           room->door[j].cdir == -1 )
         room->door[j].other = NULL;
       else
         room->door[j].other = find_door_vnum_world( room->parent->parent, room->door[j].czone, room->door[j].croom, room->door[j].cdir );
       }
}

int calc_world_size( WORLD *world )
{
    int size = 0;
    int j, k;

    size += sizeof(WORLD);
    for( j = 0; j < world->num_zones; j++ ){
       size += sizeof(ZONE);
       size += NAMESIZE*2;   /* Author, Title */
       for( k = 0; k < world->zone[j].num_rooms; k++ ){
          size += sizeof(ROOM);
          size += NAMESIZE;    /* Title */
          size += RDSIZE;      /* Description */

          size += world->zone[j].room[k].num_doors*sizeof(DOOR);
          size += world->zone[j].room[k].num_doors*NAMESIZE;  /* Title */
          size += world->zone[j].room[k].num_doors*RDSIZE;    /* Description */

          size += world->zone[j].room[k].num_exd*sizeof(EXD);
          size += world->zone[j].room[k].num_exd*NAMESIZE;  /* Title */
          size += world->zone[j].room[k].num_exd*RDSIZE;    /* Description */

          size += world->zone[j].room[k].num_cont*sizeof(RCONTENT);
          }
       for( k = 0; k < world->zone[j].num_objects; k++ ){
          size += sizeof(OBJECT);
          size += RDSIZE*4;   /* keywords, short_desc, long_desc, action_desc */

          size += world->zone[j].object[k].num_exd*sizeof(EXD);
          size += world->zone[j].object[k].num_exd*NAMESIZE;  /* Title */
          size += world->zone[j].object[k].num_exd*RDSIZE;    /* Description */

          size += world->zone[j].object[k].num_exd*sizeof(OAFFECT);
          }
       for( k = 0; k < world->zone[j].num_mobiles; k++ ){
          size += sizeof(MOBILE);
          size += RDSIZE*3;   /* keywords, short_desc, long_desc, detail_desc */

          size += world->zone[j].mobile[k].num_especs*sizeof(ESPEC);
          }
       for( k = 0; k < world->zone[j].num_shops; k++ ){
          size += sizeof(SHOP);

          size += world->zone[j].shop[k].num_items*sizeof(int);

          size += world->zone[j].shop[k].num_bspecs*sizeof(BSPEC);
          size += world->zone[j].shop[k].num_bspecs*NAMESIZE; /* namelist */

          size += NAMESIZE*7; /* message when... x7 */

          size += world->zone[j].shop[k].num_rooms*sizeof(int);
          }
       for( k = 0; k < world->zone[j].num_tags; k++ ){
          size += sizeof(TAG);
          size += NAMESIZE;   /* str */
          }
       }

    return size;
}

int zone_sorter( const void *e1, const void *e2 )
{
    return ((ZONE *)e1)->vnum - ((ZONE *)e2)->vnum;
}

void sort_zones( WORLD *world )
{
    qsort( world->zone, world->num_zones, sizeof(ZONE), zone_sorter );
}

int destroy_board_data( BOARD *board )
{
    if( !board->data )
      return 0;

    switch( board->type ){
      case CBT_ROOM:
           {
            ROOM *temp;

            temp = board->data;

            free( temp->title );
            free( temp->desc );
            if( temp->exd )
              free( temp->exd );
            if( temp->content )
              free( temp->content );
            }
           break;
      }
    free( board->data );
    board->data = NULL;

    return 0;
}

int copy_room_to_board( ROOM *room, BOARD *board )
{
    ROOM *dest;
    int j;

    board->type = CBT_ROOM;

    destroy_board_data( board );

    board->data = malloc( sizeof(ROOM) );

    dest = board->data;

    dest->x = room->x;
    dest->y = room->y;
    dest->z = room->z;

    dest->title = malloc( NAMESIZE );
    strcpy( dest->title, room->title );

    dest->desc = malloc( RDSIZE );
    strcpy( dest->desc, room->desc );

    dest->vnum = 0;

    dest->zone = room->parent->vnum;

    dest->rbv = room->rbv;
    dest->stype = room->stype;

    dest->num_doors = 0;
    dest->door = NULL;

    dest->num_exd = room->num_exd;
    if( dest->num_exd > 0 )
      dest->exd = malloc( sizeof(EXD)*dest->num_exd );
    else
      dest->exd = NULL;
    for( j = 0; j < dest->num_exd; j++ ){
       dest->exd[j].keywords = malloc( NAMESIZE );
       strcpy( dest->exd[j].keywords, room->exd[j].keywords );

       dest->exd[j].desc = malloc( RDSIZE );
       strcpy( dest->exd[j].desc, room->exd[j].desc );
       }

    dest->num_cont = room->num_cont;
    if( dest->num_cont > 0 )
      dest->content = malloc( sizeof(RCONTENT)*dest->num_cont );
    else
      dest->content = NULL;
    for( j = 0; j < dest->num_cont; j++ ){
       dest->content[j].type   = room->content[j].type;
       dest->content[j].vnum   = room->content[j].vnum;
       dest->content[j].ifflag = room->content[j].ifflag;
       dest->content[j].max    = room->content[j].max;
       dest->content[j].extra  = room->content[j].extra;
       }

    dest->parent = NULL;

    return 0;
}

int paste_board_to_zone( BOARD *board, ZONE *zone )
{
    int j, ibf=-1;

    if( !board->data )
      return 0;

    switch( board->type ){
      case CBT_ROOM:
           {
            ROOM *temp;

            temp = board->data;

            ibf = add_room( zone, temp->x, temp->y, temp->z );
            strcpy( zone->room[ibf].title, temp->title );
            strcpy( zone->room[ibf].desc, temp->desc );
            zone->room[ibf].rbv = temp->rbv;
            zone->room[ibf].stype = temp->stype;

            zone->room[ibf].num_exd = temp->num_exd;
            if( zone->room[ibf].num_exd > 0 )
              zone->room[ibf].exd = realloc( zone->room[ibf].exd, sizeof(EXD)*zone->room[ibf].num_exd );
            for( j = 0; j < zone->room[ibf].num_exd; j++ ){
               zone->room[ibf].exd[j].keywords = malloc( NAMESIZE );
               strcpy( zone->room[ibf].exd[j].keywords, temp->exd[j].keywords );
        
               zone->room[ibf].exd[j].desc = malloc( RDSIZE );
               strcpy( zone->room[ibf].exd[j].desc, temp->exd[j].desc );
               }

            if( zone->vnum == temp->zone ){
              zone->room[ibf].num_cont = temp->num_cont;
              if( zone->room[ibf].num_cont > 0 )
                zone->room[ibf].content = realloc( zone->room[ibf].content, sizeof(RCONTENT)*zone->room[ibf].num_cont );
              for( j = 0; j < zone->room[ibf].num_cont; j++ ){
                 zone->room[ibf].content[j].type   = temp->content[j].type;
                 zone->room[ibf].content[j].vnum   = temp->content[j].vnum;
                 zone->room[ibf].content[j].ifflag = temp->content[j].ifflag;
                 zone->room[ibf].content[j].max    = temp->content[j].max;
                 zone->room[ibf].content[j].extra  = temp->content[j].extra;
                 }
              }
            }
           break;
      }

    return ibf;
}

MSEL_ITEM *add_msel_item( MSEL *msel, int hi_ind, int lo_ind )
{
    MSEL_ITEM *cur, *next;

    if( msel->base == NULL || msel->num_sel < 1 ){
      msel->base = malloc( sizeof(MSEL_ITEM) );
      msel->base->hi_ind = hi_ind;
      msel->base->lo_ind = lo_ind;
      msel->base->next = msel->base;
      msel->num_sel++;
      return msel->base;
      }
    else{
      if( find_msel_item( msel, hi_ind, lo_ind ) ){
        return NULL;
        }
      next = msel->base->next;
      msel->base->next = malloc( sizeof(MSEL_ITEM) );
      cur = msel->base->next;
      cur->hi_ind = hi_ind;
      cur->lo_ind = lo_ind;
      cur->next = next;
      msel->num_sel++;
      return cur;
      }
}

int del_msel_item( MSEL *msel, int hi_ind, int lo_ind )
{
    MSEL_ITEM *last, *next;

    if( (last = find_msel_item( msel, hi_ind, lo_ind ) ) == NULL )
      return -1;

    if( last->next == last ){
      free( last );
      msel->num_sel = 0;
      msel->base = NULL;
      return 0;
      }

    next = last->next->next;
    free( last->next );
    last->next = next;
    msel->num_sel--;

    return 1;
}

int del_msel_list( MSEL *msel )
{
    MSEL_ITEM *cur, *next;

    if( (cur = msel->base) == NULL )
      return 0;

    do{
      next = cur->next;
      free( cur );
      cur = next;
      msel->num_sel--;
      } while( cur != msel->base );

    msel->base = NULL;
    if( msel->num_sel ){
      fprintf( stderr, "msel->num_sel = %d!!!\n", msel->num_sel );
      }
    msel->num_sel = 0;

    return 0;
}
